#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>

#include "gauss.h"

Gauss::Gauss(int seed)
{
  mSecond            = false;
  mMean              = 1;
  mStandardDeviation = 0;
  mMin = -1e20f;
  mMax = 1e20f;
  Rand::Set(seed); // init random number generator
};

Gauss::Gauss(const char *str,int seed)
{
  mSecond            = false;
  mMean              = 1;
  mStandardDeviation = 0;
  mMin = -1e20f;
  mMax = 1e20f;
  Rand::Set(seed); // init random number generator
  Set(str);
};

Gauss::~Gauss(void)
{
};

float Gauss::Get(void)
{
  float v = RandGauss()*mStandardDeviation + mMean;
  if ( v < mMin ) v = mMin;
  if ( v > mMax ) v = mMax;
  mCurrent = v;
  return v;
};

float Gauss::GetCurrent(void) const
{
  return mCurrent;
};

void Gauss::Set(float m,float s)
{
  mMean = m;
  mStandardDeviation = s;
};

// set directly.
void Gauss::Set(float m,float s,float min,float max)
{
  mMean = m;
  mStandardDeviation = s;
  mMin = min;
  mMax = max;
}

// Set from ascii string.
const char * Gauss::Set(const char *arg)
{
  char *end;
  strtogmd( (char *)arg, &end, mMean, mStandardDeviation, mMin, mMax );
  return end;
};

float Gauss::GetMean(void) const
{
  return mMean;
};

float Gauss::GetStandardDeviation(void) const
{
  return mStandardDeviation;
};

void Gauss::SetMean(float m)
{
  mMean = m;
};

void Gauss::SetStandardDeviation(float s)
{
  mStandardDeviation = s;
};

// convert gaussian into valid gaussian string.
void Gauss::GetString(char *scratch) const
{
  if ( mStandardDeviation != 0 || mMin > -1e20f || mMax <  1e20f )
  {
    if ( mMin > -1e20f || mMax < 1e20f )
      sprintf(scratch,"%0.3f:%0.3f<%0.3f:%0.3f>",mMean,mStandardDeviation,mMin,mMax);
    else
      sprintf(scratch,"%0.3f:%0.3f",mMean,mStandardDeviation);
  }
  else
  {
    sprintf(scratch,"%0.3f",mMean);
  }
};

// generate fresh gaussian pair or return last valid one.
float Gauss::RandGauss( void )
{
  if( mSecond )
  {
    mSecond = false;
    return mGauss2;
  }

  float x1;
  float x2;
  float w;
  do
  {
    x1 = 2.0f * GetFloat() - 1.0f;
    x2 = 2.0f * GetFloat() - 1.0f;
    w = x1 * x1 + x2 * x2;
  } while ( w >= 1.0f );

  w = sqrtf( (-2.0f * logf( w ) ) / w );

  mGauss1 = x1 * w;
  mGauss2 = x2 * w;

  mSecond = true;

  return mGauss1;
};


// convert string to gaussian number.  Return code
// indicates number of arguments found.
int Gauss::strtogmd(const char* spec,
                    char** end,
                    float& mean,
                    float& deviation,
                    float& min,
                    float& max ) const
{
  char* pos;
  min  = -1e20f;
  max  = 1e20f;
  mean = (float) strtod( (char*) spec, &pos );
  deviation = 0;

  if( pos == spec )
  {
    mean = 0;
    if( end != 0 )
    {
      *end = (char *)spec;
    }
    return 0;
  }
  if( *pos != ':' )
  {
    if( end != 0 )
    {
      *end = pos;
    }
    return 1;
  }
  char* devp = pos + 1;
  deviation = (float) strtod( devp, &pos );
  if( pos == devp )
  {
    if( end != 0 )
    {
      *end = devp - 1;
    }
    return 1;
  }

  if( *pos != '[' && *pos != '<' )
  {
    if( end != 0 )
    {
      *end = pos;
    }
    return 2;
  }

  char* minp = pos + 1;
  char* maxp;
  min = (float) strtod( minp, &maxp );
  if( *maxp != ',' && *maxp != ':' )
  {
    if( end != 0 )
    {
      *end = pos;
    }
    return 2;
  }
  ++maxp;
  char* endp;
  max = (float) strtod( maxp, &endp );
  if( *endp != ']' && *endp != '>' )
  {
    if( end != 0 )
    {
      *end = pos;
    }
    return 2;
  }
  if( minp == maxp - 1 )
  {
    min = -1e20f;
  }
  if( maxp == endp )
  {
    max = 1e20f;
  }
  if( end != 0 )
  {
    *end = endp + 1;
  }
  return 4;
};

